/*
 * Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 and
 * only version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 */

#define SECUREC_NEED_ERRNO_TYPE
#include <securec.h>
#include "bdev_internal.h"

#include "spdk/stdinc.h"
#include "spdk/bdev.h"
#include "spdk/bdev_module.h"
#include "spdk/env.h"
#include "spdk/nvme_spec.h"
#include "spdk/log.h"

#include "spdk_internal/bdev_stat.h"

pthread_mutex_t *g_io_stat_map_mutex = NULL;
/* share memory for libstorage iostat */
struct libstorage_bdev_io_stat *g_io_stat_map;
/* libstorage iostat enable or disable switch */
int32_t g_libstorage_iostat = 0;
int32_t g_polltime_threshold = 0;

void spdk_bdev_init_iostat(struct spdk_bdev_channel *ch, struct spdk_bdev *bdev,
			   struct spdk_io_channel *io_ch, struct spdk_bdev_io_stat *stat)
{
	int		i = 0;
	bool		find = false;
	uint16_t	channel_id;

	if (bdev->fn_table->get_io_channel_id) {
		channel_id = bdev->fn_table->get_io_channel_id(io_ch);
		for (i = 0; i < STAT_MAX_NUM; i++) {
			/* Reuse last record */
			if (g_io_stat_map[i].used && !strcmp(g_io_stat_map[i].bdev_name, bdev->name)
			    && g_io_stat_map[i].channel_id == channel_id) {
				stat->io_stat_id = i;
				find = true;
				g_io_stat_map[i].num_read_ops = 0;
				g_io_stat_map[i].num_write_ops = 0;
				g_io_stat_map[i].bytes_read = 0;
				g_io_stat_map[i].bytes_written = 0;
				g_io_stat_map[i].io_outstanding = 0;
				g_io_stat_map[i].read_latency_ticks = 0;
				g_io_stat_map[i].write_latency_ticks = 0;
				g_io_stat_map[i].io_ticks = 0;
				g_io_stat_map[i].poll_time_used = false;
				g_io_stat_map[i].num_poll_timeout = 0;
				break;
			}
		}
		if (!find) {
			/* Add lock when multi thread or process */
			if (pthread_mutex_lock(g_io_stat_map_mutex) == EOWNERDEAD) {
				if (pthread_mutex_consistent(g_io_stat_map_mutex)) {
					SPDK_WARNLOG("[libstorage] the iostat_map process mutex is not normal any more.\n");
				}
			}
			for (i = 0; i < STAT_MAX_NUM; i++) {
				/* Find unused record, allocate it to this channel */
				if (!g_io_stat_map[i].used) {
					g_io_stat_map[i].used = true;
					if (strncpy_s(g_io_stat_map[i].bdev_name, sizeof(g_io_stat_map[i].bdev_name), bdev->name,
						      sizeof(g_io_stat_map[i].bdev_name) - 1) != 0) {
						SPDK_ERRLOG("[libstorage] string copy failed.\n");
					}
					g_io_stat_map[i].channel_id = channel_id;
					stat->io_stat_id = i;
					find = true;
					break;
				}
			}
			pthread_mutex_unlock(g_io_stat_map_mutex);
		}
		if (!find) {
			stat->io_stat_id = -1;
			SPDK_ERRLOG("channel %u bdev %s allocate io stat memory failed.\n", channel_id, bdev->name);
		}
	} else {
		/* It is not nvme disk, can use iostat. So do not do IO statistics in libstorage. */
		stat->io_stat_id = -1;
	}
	stat->start_tsc = spdk_get_ticks();
	stat->interval_tsc = spdk_get_ticks_hz() / 10;
}

void spdk_bdev_destroy_iostat(struct spdk_bdev_channel *ch, struct spdk_bdev *bdev,
			      struct spdk_io_channel *io_ch)
{
	int		i = 0;
	uint16_t	channel_id;

	if (bdev->fn_table->get_io_channel_id) {
		channel_id = bdev->fn_table->get_io_channel_id(io_ch);
		for (i = 0; i < STAT_MAX_NUM; i++) {
			/* clear channel iostat info in share memory */
			if (g_io_stat_map[i].used && !strcmp(g_io_stat_map[i].bdev_name, bdev->name)
			    && g_io_stat_map[i].channel_id == channel_id) {
				g_io_stat_map[i].channel_id = 0;
				memset(g_io_stat_map[i].bdev_name, 0, sizeof(g_io_stat_map[i].bdev_name));
				g_io_stat_map[i].num_read_ops = 0;
				g_io_stat_map[i].num_write_ops = 0;
				g_io_stat_map[i].bytes_read = 0;
				g_io_stat_map[i].bytes_written = 0;
				g_io_stat_map[i].io_outstanding = 0;
				g_io_stat_map[i].read_latency_ticks = 0;
				g_io_stat_map[i].write_latency_ticks = 0;
				g_io_stat_map[i].io_ticks = 0;
				/* used flag set false in last avoid race in channel create */
				g_io_stat_map[i].used = false;
				g_io_stat_map[i].poll_time_used = false;
				g_io_stat_map[i].num_poll_timeout = 0;
			}
		}
	}
}

int spdk_bdev_unmap_multiblocks(struct spdk_bdev_desc *desc, struct spdk_io_channel *ch,
				void *unmap_d, uint32_t unmap_count,
				spdk_bdev_io_completion_cb cb, void *cb_arg)
{
	struct spdk_bdev *bdev = spdk_bdev_desc_get_bdev(desc);
	struct spdk_bdev_io *bdev_io = NULL;
	struct spdk_bdev_channel *channel = spdk_io_channel_get_ctx(ch);

	bdev_io = bdev_channel_get_io(channel);
	if (bdev_io == NULL) {
		return -ENOMEM;
	}

	bdev_io->internal.ch = channel;
	bdev_io->internal.desc = desc;
	bdev_io->type = SPDK_BDEV_IO_TYPE_UNMAP_BLOCKS;
	bdev_io->u.contig.buf = unmap_d;
	bdev_io->u.contig.num_blocks = unmap_count;
	bdev_io_init(bdev_io, bdev, cb_arg, cb);

	bdev_io_submit(bdev_io);
	return 0;
}

void bdev_io_stat_update(struct spdk_bdev_io *bdev_io, uint64_t tsc_diff,
			 struct spdk_bdev_io_stat *stat)
{
	switch (bdev_io->type) {
	case SPDK_BDEV_IO_TYPE_READ_NVME:
		stat->bytes_read += bdev_io->u.contig.nbytes + bdev_io->u.contig.md_len;
		stat->num_read_ops++;
		stat->read_latency_ticks += tsc_diff;
		break;
	case SPDK_BDEV_IO_TYPE_WRITE_NVME:
		stat->bytes_written += bdev_io->u.contig.nbytes + bdev_io->u.contig.md_len;
		stat->num_write_ops++;
		stat->write_latency_ticks += tsc_diff;
		break;
	case SPDK_BDEV_IO_TYPE_READV_NVME:
		stat->bytes_read += bdev_io->u.bdev.nbytes;
		stat->num_read_ops++;
		stat->read_latency_ticks += tsc_diff;
		break;
	case SPDK_BDEV_IO_TYPE_WRITEV_NVME:
		stat->bytes_written += bdev_io->u.bdev.nbytes;
		stat->num_write_ops++;
		stat->write_latency_ticks += tsc_diff;
		break;
	default:
		break;
	}
}

void bdev_update_iostat_map(struct spdk_bdev_io *bdev_io, uint64_t tsc,
			    struct spdk_bdev_io_stat *stat,
			    struct spdk_io_channel *channel, uint64_t io_outstanding)
{
	uint64_t	num_poll_timeout;

	stat->pre_ticks = stat->cur_ticks;
	stat->cur_ticks = tsc;
	stat->io_ticks += stat->cur_ticks - stat->pre_ticks;

	if (g_libstorage_iostat) {
		if ((stat->io_stat_id >= 0) && (stat->io_stat_id < STAT_MAX_NUM)) {
			g_io_stat_map[stat->io_stat_id].io_outstanding = io_outstanding;
			if (tsc > (stat->start_tsc + stat->interval_tsc)) {
				g_io_stat_map[stat->io_stat_id].num_read_ops = stat->num_read_ops;
				g_io_stat_map[stat->io_stat_id].num_write_ops = stat->num_write_ops;
				g_io_stat_map[stat->io_stat_id].bytes_read = stat->bytes_read;
				g_io_stat_map[stat->io_stat_id].bytes_written = stat->bytes_written;
				g_io_stat_map[stat->io_stat_id].read_latency_ticks = stat->read_latency_ticks;
				g_io_stat_map[stat->io_stat_id].write_latency_ticks = stat->write_latency_ticks;
				g_io_stat_map[stat->io_stat_id].io_ticks = stat->io_ticks;

				stat->start_tsc = tsc;

				if (g_polltime_threshold) {
					num_poll_timeout = bdev_io->bdev->fn_table->get_timeout_count ? \
							   bdev_io->bdev->fn_table->get_timeout_count(channel) : 0;
					g_io_stat_map[stat->io_stat_id].poll_time_used = true;
					g_io_stat_map[stat->io_stat_id].num_poll_timeout = num_poll_timeout;
				}
			}
		}
	}
}
