/*
 * Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 and
 * only version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 */

#include "spdk/stdinc.h"
#include "nvme_internal.h"

void spdk_nvme_ctrlr_set_shutdown(struct spdk_nvme_ctrlr *ctrlr, bool is_shutdown)
{
	nvme_robust_mutex_lock(&ctrlr->ctrlr_lock);
	ctrlr->is_destructed = is_shutdown;
	nvme_robust_mutex_unlock(&ctrlr->ctrlr_lock);
}

bool spdk_nvme_ctrlr_is_smart_per_namespace_supported(struct spdk_nvme_ctrlr *ctrlr)
{
	if (NULL == ctrlr) {
		SPDK_ERRLOG("spdk_nvme_ctrlr_is_smart_per_namespace_supported: Invalid Parameters!\n");
		return false;
	}

	/* check Bit 0 of Log Page Attributes(LPA),
	to find out whether the controller supports namespace basis or not. */

	if (0 == ctrlr->cdata.lpa.ns_smart) {
		SPDK_NOTICELOG("This controller does not support the SMART information on a per namespace basis.\n");
		return false;
	}

	return true;
}

static int nvme_get_log_info(struct spdk_nvme_ctrlr *ctrlr, uint8_t log_page, uint32_t nsid,
			     			 void *payload,  uint32_t payload_size)
{
	struct nvme_completion_poll_status status = {0x0};
	int ret;

	status.done = false;
	ret = spdk_nvme_ctrlr_cmd_get_log_page(ctrlr, log_page, nsid, payload, payload_size, 0,
					       nvme_completion_poll_cb, &status);
	if (ret) {
		return ret;
	}

	while (status.done == false) {
		nvme_robust_mutex_lock(&ctrlr->ctrlr_lock);
		spdk_nvme_qpair_process_completions(ctrlr->adminq, 0);
		nvme_robust_mutex_unlock(&ctrlr->ctrlr_lock);
	}
	if (spdk_nvme_cpl_is_error(&status.cpl)) {
		SPDK_ERRLOG("spdk_nvme_ctrlr_get_smart_info failed! sc[0x%x], sct[0x%x]\n",
			    status.cpl.status.sc, status.cpl.status.sct);
		return -ENXIO;
	}
	return 0;
}

int spdk_nvme_ctrlr_get_smart_info(struct spdk_nvme_ctrlr *ctrlr, uint32_t nsid,
			       				   struct spdk_nvme_health_information_page *smart_info)
{
	struct spdk_nvme_ns *ns = NULL;

	if (NULL == ctrlr || NULL == smart_info) {
		SPDK_ERRLOG("Invalid parameters!\n");
		return -EINVAL;
	}

	/* if controller does not support namespase basis, then set the nsid to 0xFFFFFFFF, and continue the process.
	and if nsid is 0, set the nsid to 0xFFFFFFFF too. */
	if (!spdk_nvme_ctrlr_is_smart_per_namespace_supported(ctrlr) || 0 == nsid) {
		nsid = SPDK_NVME_GLOBAL_NS_TAG;
	}

	/* nsid should be 0xffffffff or on a per namespace basis. */
	if (nsid != SPDK_NVME_GLOBAL_NS_TAG) {
		ns = spdk_nvme_ctrlr_get_ns(ctrlr, nsid);
		if (NULL == ns) {
			SPDK_ERRLOG("Invalid NS %u\n", nsid);
			return -EINVAL;
		}

		/* if the namespace specified is not active, set the nsid to 0xFFFFFFFF, and continue the process. */
		if (!spdk_nvme_ns_is_active(ns)) {
			SPDK_WARNLOG("NS %u is not active\n", nsid);
			nsid = SPDK_NVME_GLOBAL_NS_TAG;
		}
	}

	return nvme_get_log_info(ctrlr, SPDK_NVME_LOG_HEALTH_INFORMATION, nsid, smart_info,
				 sizeof(struct spdk_nvme_health_information_page));
}

int spdk_nvme_ctrlr_get_error_info(struct spdk_nvme_ctrlr *ctrlr, uint32_t err_entries,
			       				   struct spdk_nvme_error_information_entry *error_info)
{
	const struct spdk_nvme_ctrlr_data *cdata = NULL;
	uint32_t nsid = SPDK_NVME_GLOBAL_NS_TAG;
	int ret;

	if (NULL == ctrlr || NULL == error_info) {
		SPDK_ERRLOG("Invalid parameters!\n");
		return -EINVAL;
	}

	/* get cdata for judging the err_entries is bigger than the maximum number of entries supported or not. */
	cdata = spdk_nvme_ctrlr_get_data(ctrlr);
	if (err_entries > (cdata->elpe + 1u)) {
		/* if the parameter is bigger,  then cut it into the maximum number supported. */
		SPDK_WARNLOG("Cannot get %d error log entries, the controller only support %d errors.\n",
			     err_entries, cdata->elpe + 1);
		err_entries = cdata->elpe + 1;
	}

	ret = nvme_get_log_info(ctrlr, SPDK_NVME_LOG_ERROR, nsid, error_info,
				sizeof(struct spdk_nvme_error_information_entry) * err_entries);
	if (ret) {
		return ret;
	}
	/* return actual count of error log pages info. */
	return err_entries;
}

bool spdk_nvme_ctrlr_is_ns_manage_supported(struct spdk_nvme_ctrlr *ctrlr)
{
	return ctrlr->cdata.oacs.ns_manage != 0;
}

bool spdk_nvme_ctrlr_is_format_supported(struct spdk_nvme_ctrlr *ctrlr)
{
	return ctrlr->cdata.oacs.format != 0;
}

bool spdk_nvme_ctrlr_is_format_all_ns(struct spdk_nvme_ctrlr *ctrlr)
{
	return ctrlr->cdata.fna.format_all_ns != 0;
}

bool spdk_nvme_ctrlr_is_directive_supported(struct spdk_nvme_ctrlr *ctrlr)
{
	return ctrlr->cdata.oacs.directives != 0;
}

void spdk_nvme_ctrlr_update_unvmcap(struct spdk_nvme_ctrlr *ctrlr)
{
	int rc;
	struct nvme_completion_poll_status  status;
	struct spdk_nvme_ctrlr_data cdata;
	if (ctrlr == NULL) {
		SPDK_ERRLOG("Parameter error\n");
		return;
	}

	status.done = false;
	rc = nvme_ctrlr_cmd_identify(ctrlr, SPDK_NVME_IDENTIFY_CTRLR, 0, 0, 0, &cdata,
				     sizeof(cdata), nvme_completion_poll_cb, &status);
	if (rc != 0) {
		return;
	}

	if (nvme_wait_for_completion_robust_lock(ctrlr->adminq, &status, &ctrlr->ctrlr_lock)) {
		SPDK_ERRLOG("Failed to identify ctrlr data, cannot update unvmcap, sct[%x], sc[%x]\n",
			    status.cpl.status.sct, status.cpl.status.sc);
		return;
	}

	ctrlr->cdata.unvmcap[0] = cdata.unvmcap[0];
	ctrlr->cdata.unvmcap[1] = cdata.unvmcap[1];
}

int32_t spdk_nvme_ctrlr_identify_directives(struct spdk_nvme_ctrlr *ctrlr, uint16_t nsid, void *payload)
{
	struct nvme_completion_poll_status status;
	int32_t res;

	if (ctrlr == NULL || payload == NULL) {
		SPDK_ERRLOG("Parameter error\n");
		return -EINVAL;
	}

	if (!spdk_nvme_ctrlr_is_directive_supported(ctrlr)) {
		SPDK_WARNLOG("The controller[%s] does not support Directives.\n", ctrlr->trid.traddr);
		return -EPERM;
	}

	status.done = false;
	res = spdk_nvme_ctrlr_cmd_directive_receive(ctrlr, nsid,
			SPDK_NVME_IDENTIFY_DIRECTIVE_RECEIVE_RETURN_PARAM,
			SPDK_NVME_DIRECTIVE_TYPE_IDENTIFY, 0, payload,
			sizeof(struct spdk_nvme_ns_identify_directive_param),
			0, 0, nvme_completion_poll_cb, &status);
	if (res != 0) {
		return res;
	}

	while (status.done == false) {
		nvme_robust_mutex_lock(&ctrlr->ctrlr_lock);
		spdk_nvme_qpair_process_completions(ctrlr->adminq, 0);
		nvme_robust_mutex_unlock(&ctrlr->ctrlr_lock);
	}

	if (spdk_nvme_cpl_is_error(&status.cpl)) {
		SPDK_ERRLOG("Failed to Identify directive! sc[0x%x], sct[0x%x]\n",
			    status.cpl.status.sc, status.cpl.status.sct);
		return -ENXIO;
	}

	return 0;
}

uint16_t spdk_nvme_get_qpair_id(struct spdk_nvme_qpair *qpair)
{
	if (qpair == NULL) {
		return -1;
	}
	return qpair->id;
}
