/*
 * Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 and
 * only version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 */

#include "spdk/stdinc.h"
#include <securec.h>
#include <sys/queue.h>
#include "spdk/log.h"
#include "spdk/nvme.h"

#define PATH_LEN 4096
#define ID_LEN 16

/* nvme that fails to bind uio */
struct failed_nvme {
	char *pci_addr;
	TAILQ_ENTRY(failed_nvme) tailq;
};

/**
  * failed nvmes list, failed nvme will send a "nvme add uevent" when we bind it back to nvme driver
  * in spdk_rebind_driver, we should ignore this event or we wouldn't stop binding this nvme to uio.
  */
static TAILQ_HEAD(failed_nvme_list,
		  failed_nvme) g_failed_nvmes = TAILQ_HEAD_INITIALIZER(g_failed_nvmes);

/* get vendor id from /sys/bus/pci/devices/pci_addr/vendor
 * get device id from /sys/bus/pci/devices/pci_addr/device
 */
static int32_t get_id_from_sysfs(const char *pci_addr, const char *id_type, char *ret_id,
				 uint8_t ret_id_len)
{
	int32_t fd = -1;
	char sysfs_path[PATH_LEN];
	char tmp_id[ID_LEN] = {0};
	char *tmp = NULL;

	/* id's length is 5 byte,like XXXX'\0' */
	if (ret_id_len < 5) {
		SPDK_ERRLOG("ret_id_len is less than 5 bytes\n");
		return -1;
	}

	/* construct path in sysfs which stores i */
	if (snprintf_s(sysfs_path, PATH_LEN, PATH_LEN - 1, "/sys/bus/pci/devices/%s/%s", pci_addr,
		       id_type) > 0) {
		fd = open(sysfs_path, O_RDONLY);
	}
	if (fd < 0) {
		SPDK_ERRLOG("fail to open %s, errno(%d): %s\n", sysfs_path, errno, strerror(errno));
		return -1;
	}

	/* id in sysfs is like 0xDDDD */
	if (read(fd, tmp_id, ID_LEN - 1) <= 0) {
		SPDK_ERRLOG("fail to read id from %s, errno(%d): %s\n", sysfs_path, errno, strerror(errno));
		close(fd);
		return -1;
	}

	/* 2 means skipping prefix "0x" of id read from sysfs */
	tmp = tmp_id + 2;
	/* 4 means the value of id read from sysfs, not including prefix "0x" */
	if (snprintf_s(ret_id, ret_id_len, 4, "%s", tmp) <= 0) {
		SPDK_ERRLOG("string copy failed\n");
	}

	close(fd);
	return 0;
}

/* get ven_dev_id which combines vendor id and device id */
static int32_t get_ven_dev_id(const char *pci_addr, char *ven_dev_id, uint8_t ven_dev_id_len)
{
	char ven_id[ID_LEN], dev_id[ID_LEN];

	/* ven_dev_id combines with vendor id and device id,like "DDDD XXXX'\0'",length is 10 bytes */
	if (ven_dev_id_len < 10) {
		SPDK_ERRLOG("ven_dev_id_len is less than 10 bytes\n");
		return -1;
	}

	/* get vendor id from sysfs,format is like "DDDD" */
	if (get_id_from_sysfs(pci_addr, "vendor", ven_id, ID_LEN) < 0) {
		SPDK_ERRLOG("fail to get vendor id\n");
		return -1;
	}

	/* get device id from sysfs,format is like "XXXX" */
	if (get_id_from_sysfs(pci_addr, "device", dev_id, ID_LEN) < 0) {
		SPDK_ERRLOG("fail to get device id\n");
		return -1;
	}

	if (snprintf_s(ven_dev_id, ven_dev_id_len, ven_dev_id_len - 1, "%s %s", ven_id, dev_id) <= 0) {
		SPDK_ERRLOG("string copy failed\n");
		return -1;
	}
	return 0;
}

/* unbind driver by writing remove_id and unbind files in sysfs */
static int32_t unbind_driver(char *pci_addr, const char *ven_dev_id)
{
	char sysfs_dev_remove_id[PATH_LEN];	/* remove_id file path in sysfs */
	char sysfs_dev_unbind[PATH_LEN];	/* unbind file path in sysfs */
	int32_t remove_id_fd = -1;		/* file description of remove_id file */
	int32_t unbind_fd = -1;			/* file description of unbind file */
	int32_t ret;

	ret = snprintf_s(sysfs_dev_remove_id, PATH_LEN, PATH_LEN - 1,
			 "/sys/bus/pci/devices/%s/driver/remove_id", pci_addr);
	if (ret <= 0) {
		SPDK_ERRLOG("copy dev id failed\n");
		return -1;
	}
	ret = snprintf_s(sysfs_dev_unbind, PATH_LEN, PATH_LEN - 1,
			 "/sys/bus/pci/devices/%s/driver/unbind", pci_addr);
	if (ret <= 0) {
		SPDK_ERRLOG("copy dev unbind failed\n");
		return -1;
	}

	remove_id_fd = open(sysfs_dev_remove_id, O_WRONLY);
	if (remove_id_fd < 0) {
		SPDK_ERRLOG("fail to open %s, errno(%d): %s\n", sysfs_dev_remove_id, errno, strerror(errno));
		return -1;
	}

	(void)write(remove_id_fd, ven_dev_id, strlen(ven_dev_id) + 1);
	close(remove_id_fd);

	/* unbind driver by wrting unbind file */
	unbind_fd = open(sysfs_dev_unbind, O_WRONLY);
	if (unbind_fd < 0) {
		SPDK_ERRLOG("fail to open %s, errno(%d): %s\n", sysfs_dev_unbind, errno, strerror(errno));
		return -1;
	}

	ret = write(unbind_fd, pci_addr, strlen(pci_addr) + 1);
	if (ret < 0) {
		SPDK_ERRLOG("write %s to %s fail, errno(%d): %s\n", pci_addr, sysfs_dev_unbind, errno,
			    strerror(errno));
		close(unbind_fd);
		return -1;
	}

	close(unbind_fd);

	return 0;
}

/* bind device to new driver by writing new_id and bind files in sysfs */
static int32_t bind_driver(const char *pci_addr, const char *ven_dev_id, const char *driver_name)
{
	char sysfs_driver_new_id[PATH_LEN];	/* new_id file path in sysfs */
	char sysfs_driver_bind[PATH_LEN];	/* bind file path in sysfs */
	int32_t new_id_fd = -1;			/* file description of new_id file */
	int32_t bind_fd = -1;			/* file descriptoin of bind file */
	int rc;

	rc = snprintf_s(sysfs_driver_new_id, PATH_LEN, PATH_LEN - 1, "/sys/bus/pci/drivers/%s/new_id",
			driver_name);
	if (rc > 0) {
		rc = snprintf_s(sysfs_driver_bind, PATH_LEN, PATH_LEN - 1, "/sys/bus/pci/drivers/%s/bind",
				driver_name);
	}
	if (rc <= 0) {
		SPDK_ERRLOG("string copy failed\n");
		return -1;
	}

	/* try to bind driver by write ven_dev_id to new_id file */
	new_id_fd = open(sysfs_driver_new_id, O_WRONLY);
	if (new_id_fd < 0) {
		SPDK_ERRLOG("fail to open %s, errno(%d): %s\n", sysfs_driver_new_id, errno, strerror(errno));
		return -1;
	}

	(void)write(new_id_fd, ven_dev_id, strlen(ven_dev_id) + 1);
	close(new_id_fd);

	/* bind driver by writing pci_addr to bind file if writing new_id file failed */
	bind_fd = open(sysfs_driver_bind, O_WRONLY);
	if (bind_fd < 0) {
		SPDK_ERRLOG("fail to open %s, errno(%d): %s\n", sysfs_driver_bind, errno, strerror(errno));
		return -1;
	}

	(void)write(bind_fd, pci_addr, strlen(pci_addr) + 1);
	close(bind_fd);
	return 0;
}

int32_t spdk_rebind_driver(char *pci_addr, char *driver_name)
{
	char ven_dev_id[ID_LEN] = {0};
	struct failed_nvme *iter = NULL;
	unsigned int sleep_time = 1000;

	if (pci_addr == NULL || driver_name == NULL) {
		SPDK_ERRLOG("pci address and driver_name can't be NULL to rebind driver\n");
		return -1;
	}

	/* ignore event from binding pci back to nvme driver */
	TAILQ_FOREACH(iter, &g_failed_nvmes, tailq) {
		if (strncmp(iter->pci_addr, pci_addr, strlen(iter->pci_addr)) == 0) {
			/* oncely ignore nvme add event from binding back to nvme,so do rebind when next hotplug of this pci happen */
			TAILQ_REMOVE(&g_failed_nvmes, iter, tailq);
			free(iter->pci_addr);
			free(iter);
			SPDK_WARNLOG("ignore failed nvme %s\n", pci_addr);
			return 0;
		}
	}

	if (get_ven_dev_id(pci_addr, ven_dev_id, ID_LEN) < 0) {
		SPDK_ERRLOG("failed to get ven_dev_id\n");
		return -1;
	}

	while (unbind_driver(pci_addr, ven_dev_id) < 0) {
		usleep(sleep_time);
		sleep_time = sleep_time * 2;
		if (sleep_time > 1000000) {
			SPDK_ERRLOG("failed to unbind driver of %s\n", pci_addr);
			return -1;
		}
	}

	if (bind_driver(pci_addr, ven_dev_id, driver_name) < 0) {
		/* retry */
		if (bind_driver(pci_addr, ven_dev_id, driver_name) < 0) {
			SPDK_ERRLOG("fail to bind %s to %s\n", pci_addr, driver_name);
			/* add fialed nvme to g_failed_nvmes */
			struct failed_nvme *failed_nvme = (struct failed_nvme *)malloc(sizeof(struct failed_nvme));
			if (failed_nvme == NULL) {
				SPDK_ERRLOG("failed to malloc for failed_nvme,can't bind %s back to nvme\n", pci_addr);
				return -1;
			}
			failed_nvme->pci_addr = strdup(pci_addr);
			if (failed_nvme->pci_addr == NULL) {
				SPDK_ERRLOG("failed to malloc for failed_nvme,can't bind %s back to nvme\n", pci_addr);
				free(failed_nvme);
				return -1;
			}
			TAILQ_INSERT_TAIL(&g_failed_nvmes, failed_nvme, tailq);

			/* bind device back to nvme driver if failed to bind uio */
			bind_driver(pci_addr, ven_dev_id, "nvme");
		}
	}
	return 0;
}
