/*
 * Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 and
 * only version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 */

#define SECUREC_NEED_ERRNO_TYPE
#include <securec.h>
#include "bdev_nvme.h"

#include "spdk/json.h"
#include "spdk/likely.h"
#include "spdk/bdev_module.h"
#include "bdev_nvme.h"
#include "spdk/nvme.h"
#include "spdk/tree.h"

#include "spdk_internal/bdev_stat.h"
#include "bdev_nvme_self.h"

void bdev_update_ch_timeout(struct nvme_poll_group *group)
{
	uint64_t current_ticks = 0;
	uint64_t poll_ticks = 0;
	int64_t poll_time = 0;

	current_ticks = spdk_get_ticks();

	if (spdk_unlikely(g_polltime_threshold)) {
		if (group->save_start_ticks) {
			poll_ticks = current_ticks - group->save_start_ticks;
			poll_time = (poll_ticks * 1000ULL) / spdk_get_ticks_hz();
			if (poll_time >= g_polltime_threshold) {
				group->num_poll_timeout++;
				SPDK_NOTICELOG("group[%p] poll timeout in %ldms", group, poll_time);
			}
		}
		group->save_start_ticks = current_ticks;
	}
}

int _bdev_nvme_submit_request_self(struct nvme_bdev_channel *nbdev_ch, struct spdk_bdev_io *bdev_io)
{
	struct nvme_bdev_io *nbdev_io = (struct nvme_bdev_io *)bdev_io->driver_ctx;

	if (nbdev_io->io_path->qpair->qpair == NULL) {
		/* The device is currently resetting */
		return -1;
	}

	switch (bdev_io->type) {
	case SPDK_BDEV_IO_TYPE_READ_NVME:
		SPDK_DEBUGLOG(bdev_nvme, "read %lu lbas with offset %#lx\n", bdev_io->u.contig.num_blocks,
			      bdev_io->u.contig.offset_blocks);
		return bdev_nvme_queue_cmd_with_md(nbdev_io->io_path->nvme_ns->ns, nbdev_io->io_path->qpair->qpair,
						   bdev_io->driver_ctx, bdev_io->u.contig.buf,
						   bdev_io->u.contig.md_buf, SPDK_BDEV_IO_TYPE_READ,
						   bdev_io->u.contig.num_blocks, bdev_io->u.contig.offset_blocks);
	case SPDK_BDEV_IO_TYPE_WRITE_NVME:
		SPDK_DEBUGLOG(bdev_nvme, "write %lu lbas with offset %#lx\n", bdev_io->u.contig.num_blocks,
			      bdev_io->u.contig.offset_blocks);
		return bdev_nvme_queue_cmd_with_md(nbdev_io->io_path->nvme_ns->ns, nbdev_io->io_path->qpair->qpair,
						   bdev_io->driver_ctx, bdev_io->u.contig.buf,
						   bdev_io->u.contig.md_buf, SPDK_BDEV_IO_TYPE_WRITE,
						   bdev_io->u.contig.num_blocks, bdev_io->u.contig.offset_blocks);
	case SPDK_BDEV_IO_TYPE_READV_NVME:
		SPDK_DEBUGLOG(bdev_nvme, "readv %lu lbas with offset %#lx\n", bdev_io->u.bdev.num_blocks,
			      bdev_io->u.bdev.offset_blocks);
		return bdev_nvme_queue_cmd_v_with_md(nbdev_io->io_path->nvme_ns->ns, nbdev_io->io_path->qpair->qpair,
						     bdev_io->driver_ctx, SPDK_BDEV_IO_TYPE_READ,
						     bdev_io->u.bdev.iovs, bdev_io->u.bdev.iovcnt,
						     bdev_io->u.bdev.num_blocks, bdev_io->u.bdev.offset_blocks);
	case SPDK_BDEV_IO_TYPE_WRITEV_NVME:
		SPDK_DEBUGLOG(bdev_nvme, "writev %lu lbas with offset %#lx\n", bdev_io->u.bdev.num_blocks,
			      bdev_io->u.bdev.offset_blocks);
		return bdev_nvme_queue_cmd_v_with_md(nbdev_io->io_path->nvme_ns->ns, nbdev_io->io_path->qpair->qpair,
						     bdev_io->driver_ctx, SPDK_BDEV_IO_TYPE_WRITE,
						     bdev_io->u.bdev.iovs, bdev_io->u.bdev.iovcnt,
						     bdev_io->u.bdev.num_blocks, bdev_io->u.bdev.offset_blocks);
	case SPDK_BDEV_IO_TYPE_UNMAP_BLOCKS:
		return bdev_nvme_unmap_blocks(nbdev_io->io_path->nvme_ns->ns,
					      nbdev_io->io_path->qpair->qpair,
					      (void *)bdev_io->driver_ctx,
					      (struct spdk_nvme_dsm_range *)bdev_io->u.contig.buf,
					      bdev_io->u.contig.num_blocks);
	default:
		return -EINVAL;
	}
}

int bdev_nvme_dump_info_json_self(void *ctx, struct spdk_json_write_ctx *w)
{
	return 0;
}

int bdev_nvme_get_io_channel_id(struct spdk_io_channel *ch)
{
	uint16_t channel_id = -1;
	struct nvme_io_path *io_path = NULL;
	struct nvme_bdev_channel *bdev_ch = spdk_io_channel_get_ctx(ch);
	io_path = STAILQ_FIRST(&bdev_ch->io_path_list);
	if (io_path == NULL) {
		return channel_id;
	}
	struct spdk_nvme_qpair *qpair = io_path->qpair->qpair;
	if (qpair != NULL) {
		channel_id = spdk_nvme_get_qpair_id(qpair);
	}
	return channel_id;
}

uint64_t bdev_nvme_get_timeout_count(struct spdk_io_channel *ch)
{
	struct nvme_ctrlr_channel *ctrlr_ch;
	if (ch == NULL) {
		return 0;
	}
	ctrlr_ch = spdk_io_channel_get_ctx(ch);
	if (ctrlr_ch->qpair == NULL || ctrlr_ch->qpair->group == NULL) {
		return 0;
	}

	return ctrlr_ch->qpair->group->num_poll_timeout;
}

int32_t nvme_ctrlr_get_info(const char *ctrlName, struct nvme_ctrlr_info **ppCtrlr)
{
	uint32_t num_ctrlr = 0, i = 0;
	struct nvme_bdev_ctrlr *nbdev_ctrlr = NULL;
	struct nvme_ctrlr *nvme_ctrlr = NULL;
	struct nvme_ctrlr_info *pCtrlrInfo = NULL;
	const struct spdk_nvme_ctrlr_data *cdata = NULL;
	const struct spdk_nvme_ctrlr_opts *opts = NULL;

	struct spdk_pci_device *pci_dev = NULL;
	int rc;

	TAILQ_FOREACH(nbdev_ctrlr, &g_nvme_bdev_ctrlrs, tailq) {
		num_ctrlr++;
	}
	if (num_ctrlr == 0) {
		SPDK_NOTICELOG("No any nvme controller.\n");
		return 0;
	}
	if (ctrlName != NULL) {
		num_ctrlr = 1;
	}
	pCtrlrInfo = calloc(num_ctrlr, sizeof(struct nvme_ctrlr_info));
	if (pCtrlrInfo == NULL) {
		SPDK_ERRLOG("Failed to alloc memory for getting controller infomation.\n");
		return -1;
	}
	TAILQ_FOREACH(nbdev_ctrlr, &g_nvme_bdev_ctrlrs, tailq) {
		if (i >= num_ctrlr) { /* prevent having controllers be added or deleted */
			i++;
			continue;
		}
		if (ctrlName != NULL) {
			if (strcmp(nbdev_ctrlr->name, ctrlName) != 0) {
				continue;
			}
		}
		nvme_ctrlr = TAILQ_FIRST(&nbdev_ctrlr->ctrlrs);
		cdata = spdk_nvme_ctrlr_get_data(nvme_ctrlr->ctrlr);
		opts = spdk_nvme_ctrlr_get_opts(nvme_ctrlr->ctrlr);
		pci_dev = spdk_nvme_ctrlr_get_pci_device(nvme_ctrlr->ctrlr);
		if (pci_dev == NULL) {
			SPDK_ERRLOG("Failed to get pci device\n");
			break;
		}
		rc = strcpy_s(pCtrlrInfo[i].ctrlName, sizeof(pCtrlrInfo[i].ctrlName), nbdev_ctrlr->name);
		if (rc != 0) {
			SPDK_ERRLOG("String copy failed\n");
		}
		rc = strcpy_s(pCtrlrInfo[i].pciAddr, sizeof(pCtrlrInfo[i].pciAddr),
			      nvme_ctrlr->active_path_id->trid.traddr);
		if (rc != 0) {
			SPDK_ERRLOG("String copy failed\n");
		}

		rc = memcpy_s(pCtrlrInfo[i].sn, sizeof(pCtrlrInfo[i].sn), cdata->sn, 20);
		if (rc != 0) {
			SPDK_ERRLOG("Memory copy failed\n");
		}

		rc = memcpy_s(pCtrlrInfo[i].fr, sizeof(pCtrlrInfo[i].fr), cdata->fr, 8);
		if (rc != 0) {
			SPDK_ERRLOG("Memory copy failed\n");
		}

		rc = memcpy_s(pCtrlrInfo[i].mn, sizeof(pCtrlrInfo[i].mn), cdata->mn, 40);
		if (rc != 0) {
			SPDK_ERRLOG("Memory copy failed\n");
		}

		pCtrlrInfo[i].trtype = (uint16_t)nvme_ctrlr->active_path_id->trid.trtype;
		pCtrlrInfo[i].tnvmcap = cdata->tnvmcap[0];
		pCtrlrInfo[i].unvmcap = cdata->unvmcap[0];
		pCtrlrInfo[i].support_ns = cdata->oacs.ns_manage;
		pCtrlrInfo[i].directives = cdata->oacs.directives;
		pCtrlrInfo[i].dsm = cdata->oncs.dsm;
		pCtrlrInfo[i].max_num_ns = cdata->nn;
		pCtrlrInfo[i].num_io_queues = opts->num_io_queues;
		pCtrlrInfo[i].io_queue_size = opts->io_queue_size;
		pCtrlrInfo[i].device_id = spdk_pci_device_get_device_id(pci_dev);
		pCtrlrInfo[i].subdevice_id = spdk_pci_device_get_subdevice_id(pci_dev);
		pCtrlrInfo[i].vid = cdata->vid;
		pCtrlrInfo[i].ssvid = cdata->ssvid;
		pCtrlrInfo[i].ctrlid = cdata->cntlid;
		pCtrlrInfo[i].version = spdk_nvme_ctrlr_get_regs_vs(nvme_ctrlr->ctrlr).raw;
		i++;
		if (ctrlName != NULL) {
			break;
		}
	}
	if (i != num_ctrlr) {
		SPDK_ERRLOG("It has controller been added or deleted when fetched infomation, please try again later.\n");
		free(pCtrlrInfo);
		return -1;
	}
	*ppCtrlr = pCtrlrInfo;
	return num_ctrlr;
}

struct spdk_nvme_ctrlr *spdk_nvme_ctrlr_get_by_name(const char *ctrlname)
{
	struct nvme_ctrlr *nvme_ctrlr = NULL;

	nvme_ctrlr = nvme_ctrlr_get_by_name(ctrlname);
	if (nvme_ctrlr == NULL) {
		return NULL;
	}

	return nvme_ctrlr->ctrlr;
}

struct spdk_nvme_ctrlr *spdk_nvme_ctrlr_get_by_ctrlr(const struct nvme_bdev_ctrlr *nbdev_ctrlr)
{
	struct nvme_ctrlr *nvme_ctrlr = NULL;

	if (nbdev_ctrlr == NULL) {
		return NULL;
	}
	nvme_ctrlr = TAILQ_FIRST(&nbdev_ctrlr->ctrlrs);

	return nvme_ctrlr->ctrlr;
}

void nvme_ctrlr_clear_iostat_by_name(const char *ctrlname)
{
	int i;
	size_t size = strnlen(ctrlname, 24);

	for (i = 0; i < STAT_MAX_NUM; i++) {
		if (strncmp(g_io_stat_map[i].bdev_name, ctrlname, size) == 0) {
			if ((g_io_stat_map[i].bdev_name[size] == 'n') && isdigit(g_io_stat_map[i].bdev_name[size + 1])) {
				g_io_stat_map[i].channel_id = 0;
				(void)memset_s(g_io_stat_map[i].bdev_name, sizeof(g_io_stat_map[i].bdev_name), 0, sizeof(g_io_stat_map[i].bdev_name));
				g_io_stat_map[i].num_read_ops = 0;
				g_io_stat_map[i].num_write_ops = 0;
				g_io_stat_map[i].bytes_read = 0;
				g_io_stat_map[i].bytes_written = 0;
				g_io_stat_map[i].io_outstanding = 0;
				g_io_stat_map[i].read_latency_ticks = 0;
				g_io_stat_map[i].write_latency_ticks = 0;
				g_io_stat_map[i].io_ticks = 0;
				/* used flag set false in last avoid race in channel create */
				g_io_stat_map[i].used = false;
			}
		}
	}
}

void nvme_ctrlr_clear_iostat_all(void)
{
	struct nvme_bdev_ctrlr	*nvme_bdev_ctrlr = NULL;

	TAILQ_FOREACH(nvme_bdev_ctrlr, &g_nvme_bdev_ctrlrs, tailq) {
		nvme_ctrlr_clear_iostat_by_name(nvme_bdev_ctrlr->name);
	}
}

struct spdk_nvme_ns *bdev_nvme_get_ns(struct nvme_bdev *nbdev)
{
	struct spdk_nvme_ctrlr *ctrlr;
	if (nbdev == NULL) {
		return NULL;
	}

	ctrlr = spdk_nvme_ctrlr_get_by_ctrlr(nbdev->nbdev_ctrlr);
	if (ctrlr == NULL) {
		return NULL;
	}

	return spdk_nvme_ctrlr_get_ns(ctrlr, nbdev->nsid);
}

void bdev_nvme_update_block_by_nvme_ctrlr(struct spdk_nvme_ctrlr *ctrlr)
{
	uint32_t nsid;
	struct nvme_bdev_ctrlr *nvme_bdev_ctrlr = NULL;
	struct nvme_bdev	*nvme_bdev = NULL;
	struct spdk_nvme_ns *ns;

	pthread_mutex_lock(&g_bdev_nvme_mutex);
	TAILQ_FOREACH(nvme_bdev_ctrlr, &g_nvme_bdev_ctrlrs, tailq) {
		if (spdk_nvme_ctrlr_get_by_ctrlr(nvme_bdev_ctrlr) != ctrlr) {
			continue;
		}

		for (nsid = spdk_nvme_ctrlr_get_first_active_ns(ctrlr);
		     nsid != 0; nsid = spdk_nvme_ctrlr_get_next_active_ns(ctrlr, nsid)) {
			ns = spdk_nvme_ctrlr_get_ns(ctrlr, nsid);
			TAILQ_FOREACH(nvme_bdev, &nvme_bdev_ctrlr->bdevs, tailq) {
				if (nvme_bdev->nsid == nsid) {
					nvme_bdev->disk.blocklen = spdk_nvme_ns_get_sector_size(ns);
					nvme_bdev->disk.blockcnt = spdk_nvme_ns_get_num_sectors(ns);
				}
			}
		}

		pthread_mutex_unlock(&g_bdev_nvme_mutex);
		return;
	}
	pthread_mutex_unlock(&g_bdev_nvme_mutex);
}

static int
nvme_ns_cmp(struct nvme_ns *ns1, struct nvme_ns *ns2)
{
	return ns1->id < ns2->id ? -1 : ns1->id > ns2->id;
}

RB_GENERATE_STATIC(nvme_ns_tree, nvme_ns, node, nvme_ns_cmp);

int bdev_nvme_update_ns(struct nvme_bdev_ctrlr *nvme_bdev_ctrlr, uint32_t nsid)
{
	struct nvme_ns	*ns = NULL;
	bool ns_active = false;
	struct spdk_nvme_ctrlr *ctrlr = NULL;
	struct nvme_ctrlr *nvme_ctrlr = NULL;


	ctrlr = spdk_nvme_ctrlr_get_by_ctrlr(nvme_bdev_ctrlr);

	if (nvme_bdev_ctrlr == NULL || ctrlr == NULL || nsid > spdk_nvme_ctrlr_get_num_ns(ctrlr)) {
		SPDK_ERRLOG("Parameter error. nsid[%u], the max nsid is[%u]\n", nsid,
			    spdk_nvme_ctrlr_get_num_ns(ctrlr));
		return -1;
	}

	nvme_ctrlr = TAILQ_FIRST(&nvme_bdev_ctrlr->ctrlrs);
	if (nvme_ctrlr == NULL) {
		SPDK_ERRLOG("Can not get nvme_ctrlr, update abort.\n");
		return -1;
	}

	ns_active = spdk_nvme_ctrlr_is_active_ns(ctrlr, nsid);

	ns = nvme_ctrlr_get_ns(nvme_ctrlr, nsid);
	if (ns == NULL && ns_active) {
		ns = nvme_ns_alloc();
		if (ns == NULL) {
			SPDK_ERRLOG("Failed to allocate namespace\n");
			return -1;
		}
		SPDK_NOTICELOG("NSID %u to be added\n", nsid);
		ns->id = nsid;
		ns->ctrlr = nvme_ctrlr;
		ns->bdev = NULL;
		RB_INSERT(nvme_ns_tree, &nvme_ctrlr->namespaces, ns);
		/* add a new bdev device in this ns */
		nvme_ctrlr_populate_namespace(nvme_ctrlr, ns);
	}

	if (ns != NULL && !ns_active) {
		SPDK_NOTICELOG("NSID %u is removed\n", nsid);
		nvme_ctrlr_depopulate_namespace(nvme_ctrlr, ns);
		return 0;
	}

	return 0;
}

bool spdk_bdev_can_remove(struct nvme_bdev_ctrlr *nvme_bdev_ctrlr, uint32_t nsid)
{
	struct nvme_bdev *bdev = NULL;
	bool empty = false;

	pthread_mutex_lock(&g_bdev_nvme_mutex);
	TAILQ_FOREACH(bdev, &nvme_bdev_ctrlr->bdevs, tailq) {
		if (bdev->nsid == nsid) {
			break;
		}
	}
	pthread_mutex_unlock(&g_bdev_nvme_mutex);

	if (bdev == NULL) {
		return true;
	}

	spdk_spin_lock(&bdev->disk.internal.spinlock);
	empty = TAILQ_EMPTY(&bdev->disk.internal.open_descs);

	if (empty) {
		/* one bdev is empty, check next until all bdev is checked */
		bdev->disk.internal.ns_status = SPDK_BDEV_NS_STATUS_REMOVING;
	}
	/* means at least one bdev is used, so we just quit this process
	 and mark the status is false. */
	spdk_spin_unlock(&bdev->disk.internal.spinlock);

	return empty;
}

void spdk_bdev_set_ns_normal(struct nvme_bdev_ctrlr *nvme_bdev_ctrlr, uint32_t nsid)
{
	struct nvme_bdev *bdev = NULL;

	pthread_mutex_lock(&g_bdev_nvme_mutex);
	TAILQ_FOREACH(bdev, &nvme_bdev_ctrlr->bdevs, tailq) {
		if (bdev->nsid == nsid) {
			break;
		}
	}
	pthread_mutex_unlock(&g_bdev_nvme_mutex);

	if (bdev == NULL) {
		return;
	}

	spdk_spin_lock(&bdev->disk.internal.spinlock);
	if (bdev->disk.internal.ns_status == SPDK_BDEV_NS_STATUS_REMOVING) {
		bdev->disk.internal.ns_status = SPDK_BDEV_NS_STATUS_READY;
	}
	spdk_spin_unlock(&bdev->disk.internal.spinlock);
}

static void check_error_type(int rc, bool read, void *qpair)
{
	if (rc == -ENOMEM) {
		SPDK_NOTICELOG("%s failed: rc = %d\n", read ? "read" : "write", rc);
	} else if (rc < 0) {
		SPDK_ERRLOG("%s failed: rc = %d, qpair is %p\n", read ? "read" : "write", rc, qpair);
	}
}

int bdev_nvme_queue_cmd_with_md(struct spdk_nvme_ns *ns, struct spdk_nvme_qpair *qpair,
				void *driver_ctx,
				void *buffer, void *metadata, int direction, uint64_t lba_count, uint64_t lba)
{
	int rc;
	uint32_t io_flags = 0;
	uint8_t* bdev_io_action = (uint8_t *)driver_ctx;
	/* filter bit 0&1 of io->pi_action to get pi_action */
	uint8_t pi_action = bdev_io_action[SPDK_BDEV_IO_ACTION_PI] & 0x03;
	uint8_t dif_flag = bdev_io_action[SPDK_BDEV_IO_ACTION_PI];
	uint8_t fua = bdev_io_action[SPDK_BDEV_IO_ACTION_FUA];
	uint16_t streamId = bdev_io_action[SPDK_BDEV_IO_STREAM_ID_0];
	uint32_t pi_type;

	streamId |= ((uint16_t)bdev_io_action[SPDK_BDEV_IO_STREAM_ID_1] << 8);
	spdk_bdev_set_io_location(driver_ctx, (uint8_t)LOCAL_LIBSTORAGE_BDEV_NVME);

	if (streamId > 0) {
		io_flags |= SPDK_NVME_IO_FLAGS_STREAMS_DIRECTIVE;
	}

	if (pi_action > IO_NO_PROTECTION) {
		pi_type = spdk_nvme_ns_get_pi_type(ns);
		if (dif_flag & FLAG_PRCHK) {
			io_flags |= SPDK_NVME_IO_FLAGS_PRCHK_GUARD;
		}
		/* type3 not support ref tag */
		if (!(dif_flag & FLAG_NO_REF) && (pi_type != SPDK_NVME_FMT_NVM_PROTECTION_TYPE3)) {
			io_flags |= SPDK_NVME_IO_FLAGS_PRCHK_REFTAG;
		}
		if (pi_action == IO_HALF_WAY_PROTECTION) {
			io_flags |= SPDK_NVME_IO_FLAGS_PRACT;
		}
	}

	if (fua) {
		io_flags |= SPDK_NVME_IO_FLAGS_FORCE_UNIT_ACCESS;
	}

	if (direction == SPDK_BDEV_IO_TYPE_READ) {
		rc = spdk_nvme_ns_cmd_read_with_md(ns, qpair, buffer, metadata, lba,
						   lba_count, bdev_nvme_queued_done, driver_ctx, io_flags, 0, 0);
	} else {
		rc = spdk_nvme_ns_cmd_write_with_md(ns, qpair, buffer, metadata, lba,
						    lba_count, bdev_nvme_queued_done, driver_ctx, io_flags, 0, streamId);
	}

	check_error_type(rc, direction == SPDK_BDEV_IO_TYPE_READ, qpair);
	return rc;
}

int bdev_nvme_queue_cmd_v_with_md(struct spdk_nvme_ns *ns, struct spdk_nvme_qpair *qpair,
				  void *driver_ctx,
				  int direction, struct iovec *iov, int iovcnt, uint64_t lba_count, uint64_t lba)
{
	int rc;
	struct nvme_bdev_io *bio = NULL;
	uint32_t io_flags = 0;
	uint8_t* bdev_io_action = (uint8_t *)driver_ctx;
	/* filter bit 0&1 of io->pi_action to get pi_action */
	uint8_t pi_action = bdev_io_action[SPDK_BDEV_IO_ACTION_PI] & 0x03;
	uint8_t dif_flag = bdev_io_action[SPDK_BDEV_IO_ACTION_PI];
	uint8_t fua = bdev_io_action[SPDK_BDEV_IO_ACTION_FUA];
	uint16_t streamId = bdev_io_action[SPDK_BDEV_IO_STREAM_ID_0];
	uint32_t pi_type;

	streamId |= ((uint16_t)bdev_io_action[SPDK_BDEV_IO_STREAM_ID_1] << 8);
	spdk_bdev_set_io_location(driver_ctx, (uint8_t)LOCAL_LIBSTORAGE_BDEV_NVME);

	if (streamId > 0) {
		io_flags |= SPDK_NVME_IO_FLAGS_STREAMS_DIRECTIVE;
	}

	if (pi_action > IO_NO_PROTECTION) {
		pi_type = spdk_nvme_ns_get_pi_type(ns);
		if (dif_flag & FLAG_PRCHK) {
			io_flags |= SPDK_NVME_IO_FLAGS_PRCHK_GUARD;
		}
		/* type3 not support ref tag */
		if (!(dif_flag & FLAG_NO_REF) && (pi_type != SPDK_NVME_FMT_NVM_PROTECTION_TYPE3)) {
			io_flags |= SPDK_NVME_IO_FLAGS_PRCHK_REFTAG;
		}
		if (pi_action == IO_HALF_WAY_PROTECTION) {
			io_flags |= SPDK_NVME_IO_FLAGS_PRACT;
		}
	}

	if (fua) {
		io_flags |= SPDK_NVME_IO_FLAGS_FORCE_UNIT_ACCESS;
	}

	bio = nvme_bdev_io_update_args((struct nvme_bdev_io *)driver_ctx, iov, iovcnt);

	if (direction == SPDK_BDEV_IO_TYPE_READ) {
		rc = spdk_nvme_ns_cmd_readv(ns, qpair, lba,
					    lba_count, bdev_nvme_queued_done, bio, io_flags,
					    bdev_nvme_queued_reset_sgl, bdev_nvme_queued_next_sge);
	} else {
		rc = spdk_nvme_ns_cmd_writev(ns, qpair, lba, lba_count,
					     bdev_nvme_queued_done, bio, io_flags,
					     bdev_nvme_queued_reset_sgl, bdev_nvme_queued_next_sge);
	}

	check_error_type(rc, direction == SPDK_BDEV_IO_TYPE_READ, qpair);
	return rc;
}

struct nvme_bdev_ctrlr *bdev_nvme_get_ctrlr_by_bdev_desc(void *bdev_desc)
{
	struct spdk_bdev *bdev = spdk_bdev_desc_get_bdev(bdev_desc);
	struct nvme_bdev *nbdev = (struct nvme_bdev *)bdev->ctxt;
	if (nbdev == NULL) {
		return NULL;
	}
	return nbdev->nbdev_ctrlr;
}

int bdev_nvme_unmap_blocks(struct spdk_nvme_ns *ns, struct spdk_nvme_qpair *qpair, void *driver_ctx,
			   struct spdk_nvme_dsm_range *unmap_d, uint32_t unmap_count)
{
	uint32_t i;

	if (unmap_count == 0 || unmap_count > SPDK_NVME_DATASET_MANAGEMENT_MAX_RANGES) {
		SPDK_ERRLOG("Invalid parameter, unmap count: %u\n", unmap_count);
		return -EINVAL;
	}

	if (unmap_d == NULL) {
		return -EINVAL;
	}

	for (i = 0; i < unmap_count; i++) {
		unmap_d[i].attributes.raw = 0;
	}

	spdk_bdev_set_io_location(driver_ctx, (uint8_t)LOCAL_LIBSTORAGE_BDEV_NVME);
	return spdk_nvme_ns_cmd_dataset_management(ns, qpair,
					     SPDK_NVME_DSM_ATTR_DEALLOCATE,
					     unmap_d, unmap_count,
					     bdev_nvme_queued_done, driver_ctx);
}

void spdk_bdev_nvme_remove_cb(void *cb_ctx, void *ctrlr)
{
	remove_cb(cb_ctx, (struct spdk_nvme_ctrlr *)ctrlr);
}

void spdk_bdev_fail_ctrlr(void *cb_ctx, void *ctrlr)
{
	spdk_nvme_ctrlr_fail_hotplug((struct spdk_nvme_ctrlr *)ctrlr);
	remove_cb(cb_ctx, (struct spdk_nvme_ctrlr *)ctrlr);
}

struct spdk_nvme_ctrlr *spdk_nvme_bdev_ctrlr_get(char *pci_trid)
{
	struct nvme_bdev_ctrlr	*nvme_bdev_ctrlr = NULL;
	struct nvme_ctrlr *nvme_ctrlr;

	TAILQ_FOREACH(nvme_bdev_ctrlr, &g_nvme_bdev_ctrlrs, tailq) {
		TAILQ_FOREACH(nvme_ctrlr, &nvme_bdev_ctrlr->ctrlrs, tailq) {
			if (strcmp(nvme_ctrlr->active_path_id->trid.traddr, pci_trid) == 0) {
				return nvme_ctrlr->ctrlr;
			}
		}
	}

	return NULL;
}

int spdk_bdev_nvme_create_self(struct spdk_nvme_transport_id *trid,
			       const char *base_name,
			       const char **names, size_t *count,
			       const char *hostnqn)
{
	struct nvme_probe_ctx	*probe_ctx = NULL;

	if (nvme_ctrlr_get(trid) != NULL) {
		SPDK_ERRLOG("A controller with the trid (traddr: %s) already exists.\n", trid->traddr);
		return -1;
	}

	probe_ctx = bdev_nvme_create_probe_ctx(trid, base_name, hostnqn);
	if (probe_ctx == NULL) {
		SPDK_ERRLOG("Failed to create probe_ctx\n");
		return -1;
	}

	if (spdk_nvme_probe(trid, probe_ctx, probe_cb, attach_cb, remove_cb)) {
		SPDK_ERRLOG("Failed to probe for new devices\n");
		free(probe_ctx);
		return -1;
	}

	free(probe_ctx);
	return 0;
}
